/*
 * RTEMS/KA9Q driver for NE2000 network card in SBC5206 ISA slot
 *
 * Written by David Fiddes, D.J.Fiddes@hw.ac.uk
 * http://www.calm.hw.ac.uk/davidf/coldfire/
 *
 *  $Id:  $
 */
#include <bsp.h>
#include <rtems/error.h>
#include <ka9q/rtems_ka9q.h>
#include <ka9q/global.h>
#include <ka9q/enet.h>
#include <ka9q/iface.h>
#include <ka9q/netuser.h>
#include <ka9q/trace.h>
#include <ka9q/commands.h>
#include <ka9q/domain.h>

static const int ei_debug = 5;

/*
 * Number of NE2000s supported by this driver
 */
#define NNS8390DRIVER	1 /* we can't have more than one card in the ISA slot as there aren't enough ints to go round */

/*
 * RTEMS event used by interrupt handler to signal daemons.
 * This must *not* be the same event used by the KA9Q task synchronization.
 */
#define INTERRUPT_EVENT	RTEMS_EVENT_1

/*
 * Receive buffer size -- Allow for a full ethernet packet plus a pointer
 */
#define RBUF_SIZE	(1520 + sizeof (struct iface *))

/*
 * Hardware-specific storage
 */
struct ns8390EnetDriver {
	NS8390 *nic;
	struct iface		*iface;
	rtems_id		txWaitTid;
	uint8 stop_page;
	uint8 rx_start_page;

	/*
	 * Statistics
	 */
	unsigned long	rxInterrupts;
	unsigned long   rxError;
	unsigned long   rxBufferOverflow;
	unsigned long   rxCounterOverflow;

	unsigned long	txInterrupts;
	unsigned long   txError;
	unsigned long	txRawWait;
};

static struct ns8390EnetDriver ns8390EnetDriver[NNS8390DRIVER]; 


static void set_dma_read_at( NS8390 *ns8390, int addr, int size )
{
  /* abort any current DMA transfers */
  ns8390->_CR = NS8390_CR_STA | NS8390_CR_RD_AB;

  /* set the size of the dma transfer */
  ns8390->regs.page0.RBCR1 = (size >> 8) & 0xFF;
  ns8390->regs.page0.RBCR0 = size & 0xFF;

  /* set the destination of the dma transfer */
  ns8390->regs.page0.CRDA1 = (addr >> 8) & 0xFF;
  ns8390->regs.page0.CRDA0 = addr & 0xFF;

  /* start remote read */
  ns8390->_CR = NS8390_CR_STA | NS8390_CR_RD_RD;
}

static void set_dma_write_at( NS8390 *ns8390, int addr, int size )
{
  int dummy_read;

  /* perform a dummy dma read - tell me what 70 and 2 are !?! */
  set_dma_read_at( ns8390, 70, 2);
  
  dummy_read = ns8390->regs.page0.RBCR0; /* dummy read */

  /* abort any current DMA transfers */
  ns8390->_CR = NS8390_CR_STA | NS8390_CR_RD_AB;

  /* set the size of the dma transfer */
  ns8390->regs.page0.RBCR1 = (size >> 8) & 0xFF;
  ns8390->regs.page0.RBCR0 = size & 0xFF;

  /* set the source of the dma transfer */
  ns8390->regs.page0.CRDA1 = (addr >> 8) & 0xFF;
  ns8390->regs.page0.CRDA0 = addr & 0xFF;

  /* start remote write */
  ns8390->_CR = NS8390_CR_STA | NS8390_CR_RD_WR;

}

void read_eth_addr( NS8390 *ns8390, uint8 *hwa )
{
  int i;
  uint8 eeprom[32];
  
  /* read the ethernet address from eprom ? */

  /* abort any current DMA transfers and stop the chip */
  ns8390->_CR = NS8390_CR_RD_AB | NS8390_CR_STP;

  /* set FIFO Threshold to 8bytes */
  ns8390->regs.page0.CNTR1 = NS8390_DCR_FT_8B;

  /* Reset the dma transfer size */
  ns8390->regs.page0.RBCR0 = 0;
  ns8390->regs.page0.RBCR1 = 0;

  /* mask and clear all the chip's interrupts */
  ns8390->regs.page0.CNTR2 = 0;
  ns8390->regs.page0.ISR = 0xFFFF;

  /* Set to do an internal loopback *//* MON seems undocumented */
  ns8390->regs.page0.RSR = NS8390_RCR_MON;
  ns8390->regs.page0.CNTR0 = NS8390_TCR_LOOP_I0;

  /* set transfer size to 32 bytes(sizeof EEPROM) */
  ns8390->regs.page0.RBCR0 = 32;
  ns8390->regs.page0.RBCR1 = 0;

  /* make transfer go to 0 */
  ns8390->regs.page0.CRDA0 = 0;
  ns8390->regs.page0.CRDA1 = 0;

  /* start the transfer */
  ns8390->_CR = NS8390_CR_STA | NS8390_CR_RD_RD;

  /* load in the EEPROM data */
  for (i = 0; i < 32; ++i)
    eeprom[i] = NS8390_DATAPORT( ns8390 );

  /* suck out the Ethernet address that we need */
  for (i = 0; i < 6; i++)
    hwa[i] = eeprom[i*2];

  /* Abort any transfer and stop the chip */
  ns8390->_CR = NS8390_CR_STP | NS8390_CR_RD_AB;
}



/*
 * ns8390 interrupt handler
 */
static rtems_isr
ns8390Enet_interrupt_handler (rtems_vector_number v)
{
  uint8 isr;
  NS8390 *ns8390 = ns8390EnetDriver[0].nic;

  /* Store the interrupt status */
  isr = ns8390->regs.page0.ISR;

  /* abort if invalid isr value */
  if (isr == 0)
    return;

  /* OK receive interrupt */
  if (isr & NS8390_ISR_PRX)
  {
    /* reset ISR bit and receive the packet */
    ns8390->regs.page0.ISR = NS8390_ISR_PRX;
    ns8390EnetDriver[0].rxInterrupts++;
    rtems_event_send (ns8390EnetDriver[0].iface->rxproc, INTERRUPT_EVENT);
  }

  /* OK transmit interrupt */
  if (isr & NS8390_ISR_PTX)
  {
    ns8390->regs.page0.ISR = NS8390_ISR_PTX;
    ns8390EnetDriver[0].txInterrupts++;
    rtems_event_send (ns8390EnetDriver[0].txWaitTid, INTERRUPT_EVENT);
  }

  /* receive error */
  if (isr & NS8390_ISR_RXE)
  {
    ns8390->regs.page0.ISR = NS8390_ISR_RXE;
    ns8390EnetDriver[0].rxError++;
  }

  /* transmit error - count */
  if (isr & NS8390_ISR_TXE)
  {
    ns8390->regs.page0.ISR = NS8390_ISR_TXE;
    ns8390EnetDriver[0].txError++;
  }

  /* buffer overflow error - count and report */
  if (isr & NS8390_ISR_OVW)
  {
    ns8390->regs.page0.ISR = NS8390_ISR_OVW;
    ns8390EnetDriver[0].rxBufferOverflow++;
  }

  /* counter overflow error - report */
  if (isr & NS8390_ISR_CNT)
  {
    ns8390->regs.page0.ISR = NS8390_ISR_CNT;
    ns8390EnetDriver[0].rxCounterOverflow++;
  }

  /* remote dma complete - ignore */
  if (isr & NS8390_ISR_RDC)
    ns8390->regs.page0.ISR = NS8390_ISR_RDC;
}

/*
 * Initialize the ethernet hardware
 */
static void
ns8390Enet_initialize_hardware (struct ns8390EnetDriver *dp, int broadcastFlag)
{
  uint8 *hwaddr;
  rtems_status_code sc;
  rtems_isr_entry old_handler;
  
  NS8390 *ns8390 = dp->nic;
  
  /* Abort any transfer and stop the chip */
  ns8390->_CR = NS8390_CR_STP | NS8390_CR_RD_AB;

  /* set the Data Configuration Register defaults */
  ns8390->regs.page0.CNTR1 = NS8390_DCR_BOS | NS8390_DCR_WTS |
                             NS8390_DCR_FT_8B;

  /* zero the dma transfer size */
  ns8390->regs.page0.RBCR0 = 0;
  ns8390->regs.page0.RBCR1 = 0;

  /* set the receive config to accept broadcasts */
  ns8390->regs.page0.RSR = NS8390_RCR_AB;

  /* set the transmit config to ENDEC module loopback */
  ns8390->regs.page0.CNTR0 = NS8390_TCR_LOOP_I0;

  /* set internal boundary buffer limit */
  ns8390->regs.page0.BNRY = 70;

  /* set default internal dma address */
  ns8390->regs.page0.CLDA1 = 70;
  ns8390->regs.page0.CLDA0 = 128;

  /* clear all the chip's interrupts */
  ns8390->regs.page0.ISR = 0xFF;

  /* Enable all interrupts except for packet transmit */
  ns8390->regs.page0.CNTR2 = NS8390_IMR_CNTE |
                             NS8390_IMR_OVWE |
                             NS8390_IMR_TXEE |
                             NS8390_IMR_RXEE |
                             NS8390_IMR_PRXE;


  /* make sure we're stopped and go to page1 regs */
  ns8390->_CR = NS8390_CR_STP   |
                NS8390_CR_RD_AB |
                NS8390_CR_PAGE1;

  /* program supplied hardware address into the chip */
  hwaddr = dp->iface->hwaddr;
  ns8390->regs.page1.PAR0 = hwaddr[0];
  ns8390->regs.page1.PAR1 = hwaddr[1];
  ns8390->regs.page1.PAR2 = hwaddr[2];
  ns8390->regs.page1.PAR3 = hwaddr[3];
  ns8390->regs.page1.PAR4 = hwaddr[4];
  ns8390->regs.page1.PAR5 = hwaddr[5];

  /* set current page register */
  ns8390->regs.page1.CURR = 71;
  
  /* not sure what this does - needed by SEEQ(?) */
/*  dp->next_receive = 0x4700; */

  /* program the multicast address */
  ns8390->regs.page1.MAR0 = 0;
  ns8390->regs.page1.MAR1 = 0;
  ns8390->regs.page1.MAR2 = 0;
  ns8390->regs.page1.MAR3 = 0;
  ns8390->regs.page1.MAR4 = 0;
  ns8390->regs.page1.MAR5 = 0;
  ns8390->regs.page1.MAR6 = 0;
  ns8390->regs.page1.MAR7 = 0;


/*
 * Set up interrupts
 */
  sc = rtems_interrupt_catch(	ns8390Enet_interrupt_handler,
				NIC_IRQ_VECTOR,
				&old_handler);
  if (sc != RTEMS_SUCCESSFUL)
    rtems_panic ("Can't attach NS8390 SCC1 interrupt handler: %s\n",
  						rtems_status_text (sc));
/*	m360.scc1.sccm = 0;*/	/* No interrupts unmasked till necessary */

  sim->icr4 = MCF5206_SIM_ICR_IL(NIC_IRQ_LEVEL) |
              MCF5206_SIM_ICR_IP(NIC_IRQ_PRIORITY);
  sim->imr = sim->imr & (~MCF5206_SIM_IMR_EINT4);	/* Enable 68HC901 interrupts */
		    
  /* start up the chip */
  ns8390->_CR = NS8390_CR_STA | NS8390_CR_RD_AB;

  /* set the transmit config to normal operation */
  ns8390->regs.page0.CNTR0 = 0;

  /* set the data config register */
  ns8390->regs.page0.CNTR1 = NS8390_DCR_WTS |
                             NS8390_DCR_BOS |
                             NS8390_DCR_LS  |
                             NS8390_DCR_FT_8B;
}

/*
 * Send raw packet (caller provides header).
 * This code runs in the context of the interface transmit
 * task or in the context of the network task.
 */
static int
ns8390Enet_raw (struct iface *iface, struct mbuf **bpp)
{
#if 0
    struct ns8390EnetDriver *dp = &ns8390EnetDriver[iface->dev];
	NS8390 *ns8390 = dp->nic;
	struct mbuf *bp;
	int data_length;
	int mod_data_length;
	uint8 *p;


	/*
	 * Fill in some logging data
	 */
	iface->rawsndcnt++;
	iface->lastsent = secclock ();
	dump (iface, IF_TRACE_OUT, *bpp);

	/*
	 * It would not do to have two tasks active in the transmit
	 * loop at the same time.
	 * The blocking is simple-minded since the odds of two tasks
	 * simultaneously attempting to use this code are low.  The only
	 * way that two tasks can try to run here is:
	 *	1) Task A enters this code and ends up having to
	 *	   wait for a transmit buffer descriptor.
	 *	2) Task B  gains control and tries to transmit a packet.
	 * The RTEMS/KA9Q scheduling semaphore ensures that there
	 * are no race conditions associated with manipulating the
	 * txWaitTid variable.
	 */
	if (dp->txWaitTid) {
		dp->txRawWait++;
		while (dp->txWaitTid)
			rtems_ka9q_ppause (10);
	}

	
	/* flip through the mbufs describing the packet to work out how big it is */
	bp = *bpp;
	data_length = bp->cnt;
	while ( bp->next != NULL )
	{
		bp = bp->next;
		data_length = data_length + bp->cnt;
	}
	
	/* upsize runt packets... */
	mod_data_length = data_length;
	if ( mod_data_length < RUNT )
	mod_data_length = RUNT;

	/* make sure we transmit an even number of bytes */
	if ((mod_data_length&1) != 0)
	mod_data_length = mod_data_length + 1;

	/* start the transmit... */
	set_dma_write_at( ns8390, 16384, mod_data_length );

	/* spool each mbuf to the chip */
	bp = *bpp;
	do {
		/* output the actual data */
		p = bp->data;
		for (index = 0; index < bp->cnt; index = index + 2)
		{
			NS8390_DATAPORT(ns8390) = *(uint16*)(p + index);
		}
		
		/* if the last mbuf and it's an odd size - write the last byte */
		if ( (bp->next = NULL) & ((bp->cnt & 1)!=0) )
		{
			NS8390_DATAPORT(ns8390) = bp->data[bp->cnt] << 8;
			data_length = data_length + 1;
		}

		bp = bp->next;
	} while ( bp != NULL );
	
	/* pad out runt packets */
	index = data_length;
	while ( index < RUNT )
  	{
		NS8390_DATAPORT(ns8390) = 0;
		index = index + 2;
	} 
	
	/* set transmit start page register */
	ns8390->regs.page0.TSR = 64;

	/* set transmit byte count msb */
	ns8390->regs.page0.NCR = (mod_data_length) & 0xFF;

	/* set transmit byte count lsb */
	ns8390->regs.page0.FIFO = ((mod_data_length) >> 8) & 0xFF;
  
	/* Thunderpackets are Go! */
	ns8390->_CR = NS8390_CR_STA | NS8390_CR_TXP;
	
	/*
	 * Show that we've finished with the packet
	 */
	dp->txWaitTid = 0;
	*bpp = NULL;
#endif
	return 0;
}

/*
 * ns8390 reader task
 */
static void
ns8390Enet_rx (int dev, void *p1, void *p2)
{
#if 0
    struct iface *iface = (struct iface *)p1;
	struct ns8390EnetDriver *dp = (struct ns8390EnetDriver *)p2;
	struct mbuf *bp;
    uint8 rxing_page;
    uint8 this_frame;
    uint8 next_frame;
    uint16 current_offset;
    NS8390 *ns8390 = dp->nic;
	
	/*
	 * Allocate space for incoming packets and start reception
	 */
  bp = ambufw (RBUF_SIZE);
  bp->data += sizeof (struct iface *);

	/*
	 * Input packet handling loop
	 */
	for (;;) {
		/*
		 * Wait for packet if there's not one ready
		 */
		if ((status = rxBd->status) & M360_BD_EMPTY) {
			/*
			 * Reset `continuous-packet' count
			 */
			continuousCount = 0;

			/*
			 * Clear old events
			 */
			m360.scc1.scce = 0x8;

			/*
			 * Wait for packet
			 * Note that the buffer descriptor is checked
			 * *before* the event wait -- this catches the
			 * possibility that a packet arrived between the
			 * `if' above, and the clearing of the event register.
			 */
			while ((status = rxBd->status) & M360_BD_EMPTY) {
				/*
				 * Unmask RXF (Full frame received) event
				 */
				m360.scc1.sccm |= 0x8;

				rtems_ka9q_event_receive (INTERRUPT_EVENT,
						RTEMS_WAIT|RTEMS_EVENT_ANY,
						RTEMS_NO_TIMEOUT);
			}
		}


	    /* Get the rx page (incoming packet pointer). */
		ns8390->_CR = NS8390_CR_RD_AB | NS8390_CR_PAGE1;
		rxing_page = ns8390->page1.CURR;
		ns8390->_CR = NS8390_CR_RD_AB | NS8390_CR_PAGE0;
		
		/* Remove one frame from the ring.  Boundary is always a page behind. */
		this_frame = ns8390->page0.BNRY + 1;
		if (this_frame >= dp->stop_page)
			this_frame = dp->rx_start_page;
		
		/* Someday we'll omit the previous, iff we never get this message.
		   (There is at least one clone claimed to have a problem.)  */
		if (ei_debug > 0  &&  this_frame != dp->current_page)
			printf("%s: mismatched read page pointers %2x vs %2x.\n",
				   iface->name, this_frame, dp->current_page);
		
		if (this_frame == rxing_page)	/* Read all the frames? */
        {
          /* hang about waiting for the next rx interrupt */
          rtems_ka9q_event_receive( INTERRUPT_EVENT,
					                RTEMS_WAIT | RTEMS_EVENT_ANY,
						            RTEMS_NO_TIMEOUT );
        }
        else
        {
          /* recieve a packet */

            current_offset = this_frame << 8;



		/*
		 * Check that packet is valid
		 */
		if ((status & (M360_BD_LAST |
				M360_BD_FIRST_IN_FRAME |
				M360_BD_LONG |
				M360_BD_NONALIGNED |
				M360_BD_SHORT |
				M360_BD_CRC_ERROR |
				M360_BD_OVERRUN |
				M360_BD_COLLISION)) ==
						(M360_BD_LAST |
						M360_BD_FIRST_IN_FRAME)) {
			/*
			 * Pass the packet up the chain
			 * The mbuf count is reduced to remove
			 * the frame check sequence at the end
			 * of the packet.
			 */
			bp = dp->rxMbuf[rxBdIndex];
			bp->cnt = rxBd->length - sizeof (uint32);
			net_route (iface, &bp);

			/*
			 * Give the network code a chance to digest the
			 * packet.  This guards against a flurry of 
			 * incoming packets (usually an ARP storm) from
			 * using up all the available memory.
			 */
			if (++continuousCount >= dp->rxBdCount)
				kwait_null ();

			/*
			 * Allocate a new mbuf
			 * FIXME: It seems to me that it would be better
			 * if there were some way to limit number of mbufs
			 * in use by this interface, but I don't see any
			 * way of determining when the mbuf we pass up
			 * is freed.
			 */
			dp->rxMbuf[rxBdIndex] = bp = ambufw (RBUF_SIZE);
			bp->data += sizeof (struct iface *);
			rxBd->buffer = bp->data;
		}
		else {
			/*
			 * Something went wrong with the reception
			 */
			if (!(status & M360_BD_LAST))
				dp->rxNotLast++;
			if (!(status & M360_BD_FIRST_IN_FRAME))
				dp->rxNotFirst++;
			if (status & M360_BD_LONG)
				dp->rxGiant++;
			if (status & M360_BD_NONALIGNED)
				dp->rxNonOctet++;
			if (status & M360_BD_SHORT)
				dp->rxRunt++;
			if (status & M360_BD_CRC_ERROR)
				dp->rxBadCRC++;
			if (status & M360_BD_OVERRUN)
				dp->rxOverrun++;
			if (status & M360_BD_COLLISION)
				dp->rxCollision++;
		}

		/*
		 * Reenable the buffer descriptor
		 */
		rxBd->status = (status & (M360_BD_WRAP | M360_BD_INTERRUPT)) | M360_BD_EMPTY;

		/*
		 * Move to next buffer descriptor
		 */
		if (++rxBdIndex == dp->rxBdCount)
			rxBdIndex = 0;
	}
#endif
}

/*
 * Shut down the interface
 * FIXME: This is a pretty simple-minded routine.  It doesn't worry
 * about cleaning up mbufs, shutting down daemons, etc.
 */
static int
ns8390Enet_stop (struct iface *iface)
{
  NS8390 *ns8390 = ns8390EnetDriver[iface->dev].nic;
  int    loop; 

  /* start up the chip aborting all transfers */
  ns8390->_CR = NS8390_CR_STA | NS8390_CR_RD_AB;

  /* mask all interrupt sources */
  ns8390->regs.page0.CNTR2 = 0;

  /* clear all interrupt sources */
  ns8390->regs.page0.ISR = 0xFF;

  for (loop = 0; loop < 0x100; ++loop)
    ;

  /* stop the chip aborting all transfers */
  ns8390->_CR = NS8390_CR_STP | NS8390_CR_RD_AB;

  return 0;
}

/*
 * Show interface statistics
 */
static void
ns8390Enet_show (struct iface *iface)
{
	printf ("      Rx Interrupts:%-8lu", ns8390EnetDriver[0].rxInterrupts);
	printf ("           Rx Error:%-8lu", ns8390EnetDriver[0].rxError);
	printf ("  Buffer Overflow:%-8lu\n", ns8390EnetDriver[0].rxBufferOverflow);
	printf (" Counter Overflow:%-8lu\n", ns8390EnetDriver[0].rxCounterOverflow);

	printf ("      Tx Interrupts:%-8lu", ns8390EnetDriver[0].txInterrupts);
	printf ("           Tx Error:%-8lu", ns8390EnetDriver[0].txError);
	printf (" Raw output wait:%-8lu\n",  ns8390EnetDriver[0].txRawWait);
}

/*
 * Attach an NS8390 driver to the system
 * This is the only `extern' function in the driver.
 *
 * argv[0]: interface label, e.g., "rtems"
 * The remainder of the arguemnts are key/value pairs:
 * mtu ##                  --  maximum transmission unit, default 1500
 * broadcast y/n           -- accept or ignore broadcast packets, default yes
 * ip ###.###.###.###      -- IP address
 * ether ##:##:##:##:##:## -- Ethernet address
 * ether eeprom            -- Get Ethernet address from the ISA card's internal EEPROM
 */
int
rtems_ka9q_driver_attach (int argc, char *argv[], void *p)
{
	struct iface *iface;
	struct ns8390EnetDriver *dp;
	char *cp;
	int i;
	int argIndex;
	int broadcastFlag;
	char cbuf[30];

	/*
	 * Find a free driver
	 */
	for (i = 0 ; i < NNS8390DRIVER ; i++) {
		if (ns8390EnetDriver[i].iface == NULL)
			break;
	}
	if (i >= NNS8390DRIVER) {
		printf ("Too many NS8390 drivers.\n");
		return -1;
	}
	if (if_lookup (argv[0]) != NULL) {
		printf ("Interface %s already exists\n", argv[0]);
		return -1;
	}
	dp = &ns8390EnetDriver[i];

	/*
	 * Create an inteface descriptor
	 */
	iface = callocw (1, sizeof *iface);
	iface->name = strdup (argv[0]);

	/*
	 * Set default values
	 */
	broadcastFlag = 1;
	dp->txWaitTid = 0;
	dp->nic = (NS8390 *)0x40000300;
	iface->mtu = 1500;
	iface->addr = Ip_addr;
	iface->hwaddr = mallocw (EADDR_LEN);
	memset (iface->hwaddr, 0x08, EADDR_LEN);

	/*
	 * Parse arguments
	 */
	for (argIndex = 1 ; argIndex < (argc - 1) ; argIndex++) {
		if (strcmp ("mtu", argv[argIndex]) == 0) {
			iface->mtu = atoi (argv[++argIndex]);
		}
		else if (strcmp ("broadcast", argv[argIndex]) == 0) {
			if (*argv[++argIndex] == 'n')
				broadcastFlag = 0;
		}
		else if (strcmp ("ip", argv[argIndex]) == 0) {
			iface->addr = resolve (argv[++argIndex]);
		}
		else if (strcmp ("ether", argv[argIndex]) == 0) {
			argIndex++;
			if (strcmp (argv[argIndex], "eeprom") == 0) {
					read_eth_addr( dp->nic, iface->hwaddr );
			}
			else {
				gether (iface->hwaddr, argv[argIndex]);
			}
		}
		else {
			printf ("Argument %d (%s) is invalid.\n", argIndex, argv[argIndex]);
			return -1;
		}
	}
	printf ("Ethernet address: %s\n", pether (cbuf, iface->hwaddr));

	/*
	 * Fill in remainder of interface configuration
	 */
	iface->dev = i;
	iface->raw = ns8390Enet_raw;
	iface->stop = ns8390Enet_stop;
	iface->show = ns8390Enet_show;
	dp->iface = iface;
	setencap (iface, "Ethernet");

	/*
	 * Set up NS8390 hardware
	 */
	ns8390Enet_initialize_hardware (dp, broadcastFlag);

	/*
	 * Chain onto list of interfaces
	 */
	iface->next = Ifaces;
	Ifaces = iface;

	/*
	 * Start I/O daemons
	 */
	cp = if_name (iface, " tx");
	iface->txproc = newproc (cp, 1024, if_tx, iface->dev, iface, NULL, 0);
	free (cp);
	cp = if_name (iface, " rx");
	iface->rxproc = newproc (cp, 1024, ns8390Enet_rx, iface->dev, iface, dp, 0);
	free (cp);

	return 0;
}

/*
 * FIXME: There should be an ioctl routine to allow things like
 * enabling/disabling reception of broadcast packets.
 */
