/*
 * RTEMS/KA9Q driver for NS8390 Ethernet Chipset
 *
 * Written by David Fiddes, D.J.Fiddes@hw.ac.uk
 * http://www.calm.hw.ac.uk/davidf/coldfire/
 *
 *  $Id:  $
 */
#include <bsp.h>
#include <rtems/error.h>
#include <ka9q/rtems_ka9q.h>
#include <ka9q/global.h>
#include <ka9q/enet.h>
#include <ka9q/iface.h>
#include <ka9q/netuser.h>
#include <ka9q/trace.h>
#include <ka9q/commands.h>
#include <ka9q/domain.h>

#include <ns8390.h>


static const char *version = 
	"NS8390 RTEMS/KA9Q sub-driver A0.01 02/08/98 David Fiddes (D.J@fiddes.surfaid.org)\n";


/*
 * Number of NS8390s supported by this driver
 */
#define NUM_NS8390_DEVICES	1 /* we'll keep things very simple for now */

/*
 * RTEMS event used by interrupt handler to signal daemons.
 * This must *not* be the same event used by the KA9Q task synchronization.
 */
#define INTERRUPT_EVENT	RTEMS_EVENT_1

/*
 * Receive buffer size -- Allow for a full ethernet packet plus a pointer
 */
#define RBUF_SIZE	(1520 + sizeof (struct iface *))

/*
 * Basic Ethernet constants
 */
#define ET_MINLEN	60


/*
 * Hardware-specific storage
 */

static struct ns8390_device ns8390_devices[NUM_NS8390_DEVICES]; 


/*
 * Device functions
 */

static int
ns8390_stop (struct iface *iface)
{
	printf("Some silly fool is trying to call ns8390_stop()\n");
	return 0;
}

static int
ns8390_raw (struct iface *iface, struct mbuf **bpp)
{
	struct ns8390_device	*dp = &ns8390_devices[iface->dev];
	unsigned8				status;
	unsigned int			len;

	unsigned32              ns8390_port;

	setRegister_f           setReg;
	getRegister_f           getReg;
	getData_f				getData;

	printf("ns8390_raw() - start\n");

	/* setup the IO functions */
	ns8390_port = dp->port;
	setReg		= dp->setRegister;
	getReg		= dp->getRegister;
	getData		= dp->getData;


	/*
	 * Fill in some logging data
	 */
	iface->rawsndcnt++;
	iface->lastsent = secclock ();
	dump (iface, IF_TRACE_OUT, *bpp);
  
	/*
	 * It would not do to have two tasks active in the transmit
	 * loop at the same time.
	 * The blocking is simple-minded since the odds of two tasks
	 * simultaneously attempting to use this code are low.  The only
	 * way that two tasks can try to run here is:
	 *	1) Task A enters this code and ends up having to
	 *	   wait for a transmit buffer descriptor.
	 *	2) Task B  gains control and tries to transmit a packet.
	 * The RTEMS/KA9Q scheduling semaphore ensures that there
	 * are no race conditions associated with manipulating the
	 * txWaitTid variable.
	 */

	if (dp->txWaitTid)
	{
		dp->txRawWait++;
		while (dp->txWaitTid)
			rtems_ka9q_ppause (10);
	}
  
	if (dp->txWaitTid == 0)		
		rtems_task_ident (0, 0, &dp->txWaitTid);


	/* get the host driver to actually send the data into the packet buffer */
	dp->block_output( dp, &len, bpp, dp->tx_start_page );

	printf( "A %i byte packet has been loaded.\n", len );

	/* free up the packet buffer */
	free_p(bpp);

	/* upsize runt packets */

	if (len < ET_MINLEN) len = ET_MINLEN;


	/* Make sure we're on PAGE0 and abort any transfer */
	(*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB );

	/* Check we're not actually tranmitting */
	status = (*getReg)( ns8390_port, NS8390_PAGE0_READ_CR );
	if (status & NS8390_CR_TXP)
	{
		printf( "KAQ9Q (%s) : Transmit attempted when the transmitter is busy.\n", dp->iface->name );
		dp->txWaitTid = 0;
		return 0;
	}
	

	/* Set the packet length */
	(*setReg)( ns8390_port, NS8390_PAGE0_WRITE_TBCR0, len & 0xFF );
	(*setReg)( ns8390_port, NS8390_PAGE0_WRITE_TBCR1, (len >> 8) & 0xFF );


	/* Where is the packet to transmit in internal memory */
	(*setReg)( ns8390_port, NS8390_PAGE0_WRITE_TPSR, dp->tx_start_page );

	/* Thunderpackets are Go! */
	(*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_RD_AB | NS8390_CR_STA | NS8390_CR_TXP );

	printf( "ns8390_raw() - end\n" );
	
	/*
	 * Show that we've finished with the packet
	 */
	dp->txWaitTid = 0;
	return 0;
}

static void
ns8390_rx (int dev, void *p1, void *p2)
{
	int						silly_counter;
    struct iface			*iface = (struct iface *)p1;
	struct ns8390_device	*dp = (struct ns8390_device *)p2;
	struct mbuf				*bp;
    uint8					rxing_page;
    uint8					this_frame;
    uint8					next_frame;
    uint16					current_offset;
	int						pkt_len;

	struct ns8390_pkt_hdr	rx_frame;

	unsigned32              ns8390_port;

	setRegister_f           setReg;
	getRegister_f           getReg;
	getData_f				getData;

	printf("Some silly fool is trying to call ns8390_rx()\n");

	/* setup the IO functions */
	ns8390_port = dp->port;
	setReg		= dp->setRegister;
	getReg		= dp->getRegister;
	getData		= dp->getData;


	/*
	 * Allocate space for incoming packets and start reception
	 */
	bp = ambufw (RBUF_SIZE);
	bp->data += sizeof (struct iface *);


	silly_counter = 0;
	for ( ;; )
	{


	    /* Get the rx page (incoming packet pointer). */
		(*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE1 | NS8390_CR_RD_AB );
		rxing_page = (*getReg)( ns8390_port, NS8390_PAGE1_CURR );
		(*setReg)( ns8390_port, NS8390_PAGE1_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB );


		/* Remove one frame from the ring.  Boundary is always a page behind. */
		this_frame = (*getReg)( ns8390_port, NS8390_PAGE0_READ_BNRY ) + 1;
		if (this_frame >= dp->stop_page)
			this_frame = dp->rx_start_page;

		/* Someday we'll omit the previous, if we never get this message.
		   (There is at least one clone claimed to have a problem.)  */
		if (this_frame != dp->current_page)
			printf("%s: mismatched read page pointers %2x vs %2x.\n",
				   iface->name, this_frame, dp->current_page);

		if (this_frame == rxing_page)	/* Read all the frames? */
        {
          /* hang about waiting for the next rx interrupt */
			printf( "I'll hang about waiting for an INTERRUPT_EVENT...\n");
			rtems_ka9q_event_receive (	INTERRUPT_EVENT,
										RTEMS_WAIT|RTEMS_EVENT_ANY,
										RTEMS_NO_TIMEOUT);

			printf( "I got an INTERRUPT_EVENT!(%i)\n", silly_counter );
        }
        else
        {
			/* recieve a packet */

            current_offset = this_frame << 8;

			/* suck the header from the chip */
			dp->get_ns8390_hdr( dp, &rx_frame, this_frame );


			pkt_len = rx_frame.count - sizeof(struct ns8390_pkt_hdr);

			next_frame = this_frame + 1 + ((pkt_len+4)>>8);

			/* should really do the 3c503 "bogosity" check here... */

			if ( (pkt_len < 60) || (pkt_len > 1518) )
			{
				printf( "bogus packet size = %d", pkt_len);
				dp->eth_stats.rx_errors++;
			}
			else if ( (rx_frame.status & 0x0F) == NS8390_RSR_PRX )
			{

				dp->block_input( dp, pkt_len, bp, current_offset + sizeof(rx_frame) );

				net_route (iface, &bp);

				dp->eth_stats.rx_packets++;

				bp = ambufw (RBUF_SIZE);
				bp->data += sizeof (struct iface *);
			}
			else
			{
				printf( "bogus packet status = %#2x\n", rx_frame.status );
			}

			next_frame = rx_frame.next;

			dp->current_page = next_frame;

			(*setReg)( ns8390_port, NS8390_PAGE0_WRITE_BNRY, next_frame - 1 );
		}


		/* go round again */
		silly_counter++;
	}


	/* we never get here! */
}

static void
ns8390_show (struct iface *iface)
{
	printf("Some silly fool is trying to call ns8390_show()\n");
	return;
}


/* External functions called by ns8390.c users */

int
ns8390_driver_attach ( struct ns8390_device **user_dp )
{
	struct iface *iface;
	struct ns8390_device *dp;
	int device_number;

	/* Print module version information */
	printf( version );

	/*
	 * Find a free driver
	 */
	for (device_number = 0 ; device_number < NUM_NS8390_DEVICES ; device_number++) {
		if (ns8390_devices[device_number].iface == NULL)
			break;
	}
	if (device_number >= NUM_NS8390_DEVICES) {
		printf ("Too many NS8390 based drivers.\n");
		return -1;
	}

	dp = &ns8390_devices[device_number];
	*user_dp = dp;

	/* Zero out the device structure */
	memset( dp, 0, sizeof(struct ns8390_device) );


	/*
	 * Set default values
	 */
	dp->txWaitTid = 0;


	/*
	 * Create an inteface descriptor
	 */
	iface = callocw (1, sizeof *iface);
	dp->iface = iface;

	/*
	 * Fill in the interface configuration
	 */
	iface->dev  = device_number;
	iface->raw  = ns8390_raw;
	iface->stop = ns8390_stop;
	iface->show = ns8390_show; 
	iface->mtu  = 1500;
	iface->addr = Ip_addr;

	setencap (iface, "Ethernet");
	iface->hwaddr = mallocw (EADDR_LEN);
	memset (iface->hwaddr, 0x08, EADDR_LEN);



	printf("ns8390_driver_attach() - end\n");

	return 0;
}


rtems_isr
ns8390_interrupt_handler (rtems_vector_number v)
{
	struct ns8390_device	*dp = &ns8390_devices[0];
	unsigned8				status;

	unsigned32              ns8390_port;

	setRegister_f           setReg;
	getRegister_f           getReg;
	getData_f				getData;

	/* setup the IO functions */
	ns8390_port = dp->port;
	setReg		= dp->setRegister;
	getReg		= dp->getRegister;
	getData		= dp->getData;

	/* Store the interrupt status */
	status = (*getReg)( ns8390_port, NS8390_PAGE0_READ_ISR );

	/* abort if invalid isr value */
	if (status == 0)
		return;

	/* OK receive interrupt */
	if (status & NS8390_ISR_PRX)
	{
		/* reset ISR bit and receive the packet */
		(*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_PRX );
		dp->eth_stats.rx_interrupts++;
		rtems_event_send (dp->iface->rxproc, INTERRUPT_EVENT);
	}

	/* OK transmit interrupt */
	if (status & NS8390_ISR_PTX)
	{
		/* reset ISR bit and allow further packets to be transmitted */
		(*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_PTX );
		dp->eth_stats.tx_interrupts++;
		rtems_event_send (dp->txWaitTid, INTERRUPT_EVENT);
	}

	/* receive error  - count */
	if (status & NS8390_ISR_RXE)
	{
		(*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_RXE );
		dp->eth_stats.rx_errors++;
	}

	/* transmit error - count */
	if (status & NS8390_ISR_TXE)
	{
		(*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_TXE );
		dp->eth_stats.tx_errors++;
	}

	/* buffer overflow error - count and report */
	if (status & NS8390_ISR_OVW)
	{
		(*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_OVW );
		dp->eth_stats.rx_over_errors++;
	}

	/* counter overflow error - report */
	if (status & NS8390_ISR_CNT)
	{
   		(*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_CNT );
		dp->eth_stats.rx_frame_errors++;
	}

	/* remote dma complete - ignore */
	if (status & NS8390_ISR_RDC)
   		(*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_RDC );
}


/*
 * Initialize the ethernet hardware
 */
void
ns8390_init( struct ns8390_device *dp )
{
	uint8					*hwaddr;
	unsigned32              ne2000_port;

	setRegister_f           setReg;
	getRegister_f           getReg;
	getData_f				getData;


	printf( "ns8390_init() - start\n" );

	/* setup the IO functions */
	ne2000_port = dp->port;
	setReg		= dp->setRegister;
	getReg		= dp->getRegister;
	getData		= dp->getData;
  
	/* Make sure we're on PAGE0, abort any transfer and stop the chip */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_STP | NS8390_CR_RD_AB );
			
	/* set the Data Configuration Register defaults */
  	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_DCR, NS8390_DCR_BOS | NS8390_DCR_WTS | NS8390_DCR_FT_8B );

	/* zero the dma transfer size */
  	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR0, 0 );
  	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR1, 0 );

	/* set the receive config to monitor */
  	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RCR, NS8390_RCR_MON );

	/* set the transmit config to ENDEC module loopback */
  	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_TCR, NS8390_TCR_LOOP_I0 );

	/* set the transmit start page */
  	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_TPSR, dp->tx_start_page );

	/* zero the tx packet buffer sizes */
	dp->tx1 = 0;
	dp->tx2 = 0;

	/* set start and stop receive addresses */
  	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_PSTART, dp->rx_start_page );
  	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_PSTOP, dp->stop_page );

	/* set internal boundary buffer limit */
  	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_BNRY, dp->stop_page - 1 );

	/* set the current packet pointer to the start of the region */
	dp->current_page = dp->rx_start_page;

	/* mask all the chip's interrupts */
  	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_IMR, 0x00 );

	/* disable all interrupts */
  	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_ISR, 0xFF );

	/* Go to PAGE1, abort any transfer and stop the chip */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE1 | NS8390_CR_STP | NS8390_CR_RD_AB );


	/* program supplied hardware address into the chip */
	hwaddr = dp->iface->hwaddr;
	(*setReg)( ne2000_port, NS8390_PAGE1_PAR0, hwaddr[0] );
	(*setReg)( ne2000_port, NS8390_PAGE1_PAR1, hwaddr[1] );
	(*setReg)( ne2000_port, NS8390_PAGE1_PAR2, hwaddr[2] );
	(*setReg)( ne2000_port, NS8390_PAGE1_PAR3, hwaddr[3] );
	(*setReg)( ne2000_port, NS8390_PAGE1_PAR4, hwaddr[4] );
	(*setReg)( ne2000_port, NS8390_PAGE1_PAR5, hwaddr[5] );


	/* program the multicast address  - to zero because we don't do multicast */
	(*setReg)( ne2000_port, NS8390_PAGE1_MAR0, 0 );
	(*setReg)( ne2000_port, NS8390_PAGE1_MAR1, 0 );
	(*setReg)( ne2000_port, NS8390_PAGE1_MAR2, 0 );
	(*setReg)( ne2000_port, NS8390_PAGE1_MAR3, 0 );
	(*setReg)( ne2000_port, NS8390_PAGE1_MAR4, 0 );
	(*setReg)( ne2000_port, NS8390_PAGE1_MAR5, 0 );
	(*setReg)( ne2000_port, NS8390_PAGE1_MAR6, 0 );
	(*setReg)( ne2000_port, NS8390_PAGE1_MAR7, 0 );

	/* set current page register */
	(*setReg)( ne2000_port, NS8390_PAGE1_CURR, dp->rx_start_page );

	/* Return to PAGE0, abort any transfer and stop the chip */
	(*setReg)( ne2000_port, NS8390_PAGE1_CR, NS8390_CR_PAGE0 | NS8390_CR_STP | NS8390_CR_RD_AB );


	printf( "ns8390_init() - end\n" );
}


void
ns8390_start( struct ns8390_device *dp )
{
	char					*cp;
	struct iface			*iface;
	unsigned32              ne2000_port;

	setRegister_f           setReg;
	getRegister_f           getReg;


	printf("ns8390_start() - start\n");

	/* setup the IO functions */
	ne2000_port = dp->port;
	setReg		= dp->setRegister;
	getReg		= dp->getRegister;
	

	/* Final NS8390 configuration and startup */

	/* disable all interrupts */
  	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_ISR, 0xFF );

	/* unmask all interrupts */
  	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_IMR,	NS8390_IMR_PRXE |
													NS8390_IMR_PTXE |
													NS8390_IMR_RXEE |
													NS8390_IMR_TXEE |
													NS8390_IMR_OVWE |
													NS8390_IMR_CNTE   );

	/* Make sure we're on PAGE0, abort any transfer and startup the chip */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_STA | NS8390_CR_RD_AB );

	/* set the transmit config for normal operation */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_TCR, 0 );

	/* set the receive config for normal operation */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RCR, NS8390_RCR_AB );

	
	iface = dp->iface;

 	/*
	 * Chain onto list of interfaces
	 */
	iface->next = Ifaces;
	Ifaces = iface;

	/*
	 * Start I/O daemons
	 */
	cp = if_name (iface, " tx");
	iface->txproc = newproc (cp, 1024, if_tx, iface->dev, iface, NULL, 0);
	free (cp);
	cp = if_name (iface, " rx");
	iface->rxproc = newproc (cp, 1024, ns8390_rx, iface->dev, iface, dp, 0);
	free (cp); 

	printf("ns8390_start() - end\n");
}